"""
Array of intervals corresponding to IntervalArray of Pandas.
Used for IntervalIndex, which is necessary for Series.value_counts() with 'bins'
argument.
"""
import numba
import pandas as pd
from numba.core import cgutils, types
from numba.extending import NativeValue, box, intrinsic, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_model, typeof_impl, unbox
from numba.parfors.array_analysis import ArrayAnalysis
import bodo


class IntervalType(types.Type):

    def __init__(self):
        super(IntervalType, self).__init__('IntervalType()')


class IntervalArrayType(types.ArrayCompatible):

    def __init__(self, arr_type):
        self.arr_type = arr_type
        self.dtype = IntervalType()
        super(IntervalArrayType, self).__init__(name=
            f'IntervalArrayType({arr_type})')

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    def copy(self):
        return IntervalArrayType(self.arr_type)


@register_model(IntervalArrayType)
class IntervalArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        lvbk__ajqe = [('left', fe_type.arr_type), ('right', fe_type.arr_type)]
        models.StructModel.__init__(self, dmm, fe_type, lvbk__ajqe)


make_attribute_wrapper(IntervalArrayType, 'left', '_left')
make_attribute_wrapper(IntervalArrayType, 'right', '_right')


@typeof_impl.register(pd.arrays.IntervalArray)
def typeof_interval_array(val, c):
    arr_type = bodo.typeof(val._left)
    return IntervalArrayType(arr_type)


@intrinsic
def init_interval_array(typingctx, left, right=None):
    assert left == right, 'Interval left/right array types should be the same'

    def codegen(context, builder, signature, args):
        avm__jzgoe, xgtz__rfgl = args
        axaci__zen = cgutils.create_struct_proxy(signature.return_type)(context
            , builder)
        axaci__zen.left = avm__jzgoe
        axaci__zen.right = xgtz__rfgl
        context.nrt.incref(builder, signature.args[0], avm__jzgoe)
        context.nrt.incref(builder, signature.args[1], xgtz__rfgl)
        return axaci__zen._getvalue()
    qjghc__vaiw = IntervalArrayType(left)
    hpys__effz = qjghc__vaiw(left, right)
    return hpys__effz, codegen


def init_interval_array_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 2 and not kws
    ubd__nwon = []
    for kzdxy__itdkt in args:
        xnyr__iqgep = equiv_set.get_shape(kzdxy__itdkt)
        if xnyr__iqgep is not None:
            ubd__nwon.append(xnyr__iqgep[0])
    if len(ubd__nwon) > 1:
        equiv_set.insert_equiv(*ubd__nwon)
    left = args[0]
    if equiv_set.has_shape(left):
        return ArrayAnalysis.AnalyzeResult(shape=left, pre=[])
    return None


(ArrayAnalysis._analyze_op_call_bodo_libs_interval_arr_ext_init_interval_array
    ) = init_interval_array_equiv


def alias_ext_init_interval_array(lhs_name, args, alias_map, arg_aliases):
    assert len(args) == 2
    numba.core.ir_utils._add_alias(lhs_name, args[0].name, alias_map,
        arg_aliases)
    numba.core.ir_utils._add_alias(lhs_name, args[1].name, alias_map,
        arg_aliases)


numba.core.ir_utils.alias_func_extensions['init_interval_array',
    'bodo.libs.int_arr_ext'] = alias_ext_init_interval_array


@box(IntervalArrayType)
def box_interval_arr(typ, val, c):
    axaci__zen = cgutils.create_struct_proxy(typ)(c.context, c.builder, val)
    c.context.nrt.incref(c.builder, typ.arr_type, axaci__zen.left)
    imlyq__nhw = c.pyapi.from_native_value(typ.arr_type, axaci__zen.left, c
        .env_manager)
    c.context.nrt.incref(c.builder, typ.arr_type, axaci__zen.right)
    kyh__bpa = c.pyapi.from_native_value(typ.arr_type, axaci__zen.right, c.
        env_manager)
    flnz__znkkn = c.context.insert_const_string(c.builder.module, 'pandas')
    lpnm__yte = c.pyapi.import_module_noblock(flnz__znkkn)
    zttsz__zmu = c.pyapi.object_getattr_string(lpnm__yte, 'arrays')
    oaqn__knvt = c.pyapi.object_getattr_string(zttsz__zmu, 'IntervalArray')
    oevf__uihn = c.pyapi.call_method(oaqn__knvt, 'from_arrays', (imlyq__nhw,
        kyh__bpa))
    c.pyapi.decref(imlyq__nhw)
    c.pyapi.decref(kyh__bpa)
    c.pyapi.decref(lpnm__yte)
    c.pyapi.decref(zttsz__zmu)
    c.pyapi.decref(oaqn__knvt)
    c.context.nrt.decref(c.builder, typ, val)
    return oevf__uihn


@unbox(IntervalArrayType)
def unbox_interval_arr(typ, val, c):
    imlyq__nhw = c.pyapi.object_getattr_string(val, '_left')
    left = c.pyapi.to_native_value(typ.arr_type, imlyq__nhw).value
    c.pyapi.decref(imlyq__nhw)
    kyh__bpa = c.pyapi.object_getattr_string(val, '_right')
    right = c.pyapi.to_native_value(typ.arr_type, kyh__bpa).value
    c.pyapi.decref(kyh__bpa)
    axaci__zen = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    axaci__zen.left = left
    axaci__zen.right = right
    olqqw__qcm = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(axaci__zen._getvalue(), is_error=olqqw__qcm)


@overload(len, no_unliteral=True)
def overload_interval_arr_len(A):
    if isinstance(A, IntervalArrayType):
        return lambda A: len(A._left)


@overload_attribute(IntervalArrayType, 'shape')
def overload_interval_arr_shape(A):
    return lambda A: (len(A._left),)


@overload_attribute(IntervalArrayType, 'ndim')
def overload_interval_arr_ndim(A):
    return lambda A: 1


@overload_attribute(IntervalArrayType, 'nbytes')
def overload_interval_arr_nbytes(A):
    return lambda A: A._left.nbytes + A._right.nbytes


@overload_method(IntervalArrayType, 'copy', no_unliteral=True)
def overload_interval_arr_copy(A):
    return lambda A: bodo.libs.interval_arr_ext.init_interval_array(A._left
        .copy(), A._right.copy())
